@@ -1,7 +1,18 @@
 import numpy as np
 
-from typing import *
-from cftool.misc import *
+from typing import Any
+from typing import Dict
+from typing import List
+from typing import Type
+from typing import Tuple
+from typing import Union
+from typing import Callable
+from typing import Optional
+from cftool.misc import register_core
+from cftool.misc import shallow_copy_dict
+from cftool.misc import get_unique_indices
+from cftool.misc import Sampler
+from cftool.misc import LoggingMixin
 
 from abc import *
 from .misc import *
@@ -32,7 +43,6 @@ class KFold:
     >>> import numpy as np
     >>>
     >>> from cfdata.types import np_int_type
-    >>> from cfdata.tabular.misc import TaskTypes
     >>> from cfdata.tabular.toolkit import KFold
     >>> from cfdata.tabular.wrapper import TabularDataset
     >>>
@@ -40,7 +50,7 @@ class KFold:
     >>> # create an imbalance dataset
     >>> y = np.zeros(6, np_int_type)
     >>> y[[-1, -2]] = 1
-    >>> dataset = TabularDataset.from_xy(x, y, TaskTypes.CLASSIFICATION)
+    >>> dataset = TabularDataset.from_xy(x, y, "clf")
     >>> k_fold = KFold(3, dataset)
     >>> for train_fold, test_fold in k_fold:
     >>>     print(np.vstack([train_fold.dataset.x, test_fold.dataset.x]))
@@ -48,19 +58,21 @@ class KFold:
 
     """
 
-    def __init__(self,
-                 k: int,
-                 dataset: TabularDataset,
-                 **kwargs):
+    def __init__(self, k: int, dataset: TabularDataset, **kwargs: Any):
         if k <= 1:
             raise ValueError("k should be larger than 1 in KFold")
-        ratio = 1. / k
+        ratio = 1.0 / k
         self.n_list = (k - 1) * [ratio]
         self.splitter = DataSplitter(**kwargs).fit(dataset)
-        self.split_results = self._order = self._cursor = None
+        self._cursor: int
+        self._order: np.ndarray
+        self.split_results: List[SplitResult]
 
-    def __iter__(self):
-        self.split_results = self.splitter.split_multiple(self.n_list, return_remained=True)
+    def __iter__(self) -> "KFold":
+        self.split_results = self.splitter.split_multiple(
+            self.n_list,
+            return_remained=True,
+        )
         self._order = np.random.permutation(len(self.split_results)).tolist()
         self._cursor = 0
         return self
@@ -97,7 +109,6 @@ class KRandom:
     >>> import numpy as np
     >>>
     >>> from cfdata.types import np_int_type
-    >>> from cfdata.tabular.misc import TaskTypes
     >>> from cfdata.tabular.toolkit import KRandom
     >>> from cfdata.tabular.wrapper import TabularDataset
     >>>
@@ -105,7 +116,7 @@ class KRandom:
     >>> # create an imbalance dataset
     >>> y = np.zeros(6, np_int_type)
     >>> y[[-1, -2]] = 1
-    >>> dataset = TabularDataset.from_xy(x, y, TaskTypes.CLASSIFICATION)
+    >>> dataset = TabularDataset.from_xy(x, y, "clf")
     >>> k_random = KRandom(3, 2, dataset)
     >>> for train_fold, test_fold in k_random:
     >>>     print(np.vstack([train_fold.dataset.x, test_fold.dataset.x]))
@@ -113,16 +124,18 @@ class KRandom:
 
     """
 
-    def __init__(self,
-                 k: int,
-                 num_test: Union[int, float],
-                 dataset: TabularDataset,
-                 **kwargs):
-        self._cursor = None
+    def __init__(
+        self,
+        k: int,
+        num_test: Union[int, float],
+        dataset: TabularDataset,
+        **kwargs: Any,
+    ):
+        self._cursor: int
         self.k, self.num_test = k, num_test
         self.splitter = DataSplitter(**kwargs).fit(dataset)
 
-    def __iter__(self):
+    def __iter__(self) -> "KRandom":
         self._cursor = 0
         return self
 
@@ -131,7 +144,10 @@ class KRandom:
             raise StopIteration
         self._cursor += 1
         self.splitter.reset()
-        test_result, train_result = self.splitter.split_multiple([self.num_test], return_remained=True)
+        test_result, train_result = self.splitter.split_multiple(
+            [self.num_test],
+            return_remained=True,
+        )
         return train_result, test_result
 
 
@@ -159,7 +175,6 @@ class KBootstrap:
     >>> import numpy as np
     >>>
     >>> from cfdata.types import np_int_type
-    >>> from cfdata.tabular.misc import TaskTypes
     >>> from cfdata.tabular.toolkit import KBootstrap
     >>> from cfdata.tabular.wrapper import TabularDataset
     >>>
@@ -167,7 +182,7 @@ class KBootstrap:
     >>> # create an imbalance dataset
     >>> y = np.zeros(6, np_int_type)
     >>> y[[-1, -2]] = 1
-    >>> dataset = TabularDataset.from_xy(x, y, TaskTypes.CLASSIFICATION)
+    >>> dataset = TabularDataset.from_xy(x, y, "clf")
     >>> k_bootstrap = KBootstrap(3, 2, dataset)
     >>> for train_fold, test_fold in k_bootstrap:
     >>>     print(np.vstack([train_fold.dataset.x, test_fold.dataset.x]))
@@ -175,12 +190,14 @@ class KBootstrap:
 
     """
 
-    def __init__(self,
-                 k: int,
-                 num_test: Union[int, float],
-                 dataset: TabularDataset,
-                 **kwargs):
-        self._cursor = None
+    def __init__(
+        self,
+        k: int,
+        num_test: Union[int, float],
+        dataset: TabularDataset,
+        **kwargs: Any,
+    ):
+        self._cursor: int
         self.dataset = dataset
         self.num_samples = len(dataset)
         if isinstance(num_test, float):
@@ -188,7 +205,7 @@ class KBootstrap:
         self.k, self.num_test = k, num_test
         self.splitter = DataSplitter(**kwargs).fit(dataset)
 
-    def __iter__(self):
+    def __iter__(self) -> "KBootstrap":
         self._cursor = 0
         return self
 
@@ -197,7 +214,10 @@ class KBootstrap:
             raise StopIteration
         self._cursor += 1
         self.splitter.reset()
-        test_result, train_result = self.splitter.split_multiple([self.num_test], return_remained=True)
+        test_result, train_result = self.splitter.split_multiple(
+            [self.num_test],
+            return_remained=True,
+        )
         tr_indices = train_result.corresponding_indices
         tr_indices = np.random.choice(tr_indices, len(tr_indices))
         tr_set = self.dataset.split_with(tr_indices)
@@ -243,16 +263,19 @@ class ImbalancedSampler(LoggingMixin):
 
     """
 
-    def __init__(self,
-                 data: TabularData,
-                 imbalance_threshold: float = 0.1,
-                 *,
-                 shuffle: bool = True,
-                 aggregation: str = "continuous",
-                 aggregation_config: Dict[str, Any] = None,
-                 sample_method: str = "multinomial",
-                 verbose_imbalance: bool = True,
-                 verbose_level: int = 2):
+    def __init__(
+        self,
+        data: TabularData,
+        imbalance_threshold: float = 0.1,
+        *,
+        shuffle: bool = True,
+        aggregation: str = "continuous",
+        aggregation_config: Optional[Dict[str, Any]] = None,
+        sample_weights: Optional[np.ndarray] = None,
+        sample_method: str = "multinomial",
+        verbose_imbalance: bool = True,
+        verbose_level: int = 2,
+    ):
         self.data = data
         self.shuffle = shuffle
         self.imbalance_threshold = imbalance_threshold
@@ -265,35 +288,65 @@ class ImbalancedSampler(LoggingMixin):
         else:
             if aggregation_config is None:
                 aggregation_config = {}
-            self.aggregation = aggregation_dict[aggregation](data, aggregation_config, data._verbose_level)
+            base = aggregation_dict[aggregation]
+            self.aggregation = base(data, aggregation_config, data._verbose_level)
             self._num_samples = len(self.aggregation.indices2id)
-        label_recognizer = data.recognizers[-1]
-        if not self.shuffle or data.is_reg:
-            label_counts = self._label_ratios = self._sampler = None
+        if sample_weights is not None:
+            label_counts = None
+            self.sample_weights = sample_weights.copy()
+            self.sample_weights /= self.sample_weights.sum() + 1e-8
+            self._sampler = Sampler(sample_method, self.sample_weights)
         else:
-            label_counter = label_recognizer.counter
-            transform_dict = label_recognizer.transform_dict
-            label_counter = {transform_dict[k]: v for k, v in label_counter.items()}
-            label_counts = np.array([label_counter[k] for k in sorted(label_counter)], np_float_type)
-            self._label_ratios, max_label_count = label_counts / self._num_samples, label_counts.max()
-            if label_counts.min() / max_label_count >= imbalance_threshold:
-                self._sampler = None
+            self.sample_weights = None
+            if not self.shuffle or data.is_reg:
+                label_counts = self._label_ratios = self._sampler = None
             else:
-                labels = data.processed.y.ravel()
-                sample_weights = np.zeros(self._num_samples, np_float_type)
-                for i, count in enumerate(label_counts):
-                    sample_weights[labels == i] = max_label_count / count
-                sample_weights /= sample_weights.sum()
-                self._sampler = Sampler(sample_method, sample_weights)
+                label_recognizer = data.recognizers[-1]
+                if label_recognizer is None:
+                    raise ValueError(
+                        "`data` should contain label recognizer "
+                        "for `ImbalancedSampler`"
+                    )
+                label_counter = label_recognizer.counter
+                transform_dict = label_recognizer.transform_dict
+                new_counter = {transform_dict[k]: v for k, v in label_counter.items()}
+                counts_list = [new_counter[k] for k in sorted(new_counter)]
+                label_counts = np.array(counts_list, np_float_type)
+                self._label_ratios = label_counts / self._num_samples
+                max_label_count = label_counts.max()
+                if label_counts.min() / max_label_count >= imbalance_threshold:
+                    self._sampler = None
+                else:
+                    processed = data.processed
+                    if processed is None:
+                        raise ValueError(
+                            "`data` should contain `processed` "
+                            "for `ImbalancedSampler`"
+                        )
+                    if not isinstance(processed.y, np.ndarray):
+                        raise ValueError(
+                            "`data` should contain `processed.y` "
+                            "for `ImbalancedSampler`"
+                        )
+                    labels = processed.y.ravel()
+                    sample_weights = np.zeros(self._num_samples, np_float_type)
+                    for i, count in enumerate(label_counts):
+                        sample_weights[labels == i] = max_label_count / count
+                    sample_weights /= sample_weights.sum() + 1e-8
+                    self._sampler = Sampler(sample_method, sample_weights)
 
-        self._sample_method, self._verbose_level = sample_method, verbose_level
-        if verbose_imbalance and self._sampler is not None:
-            self.log_msg(
-                f"using imbalanced sampler with label counts = {label_counts.tolist()}",
-                self.info_prefix, 2
-            )
+        self._sample_method = sample_method
+        self._verbose_level = verbose_level
+        if label_counts is not None and verbose_imbalance:
+            if self._sampler is not None:
+                self.log_msg(
+                    "using imbalanced sampler with "
+                    f"label counts = {label_counts.tolist()}",
+                    self.info_prefix,
+                    2,
+                )
 
-    def __len__(self):
+    def __len__(self) -> int:
         return self._num_samples
 
     @property
@@ -305,7 +358,7 @@ class ImbalancedSampler(LoggingMixin):
         return self._sample_imbalance_flag
 
     @property
-    def label_ratios(self) -> Union[None, np.ndarray]:
+    def label_ratios(self) -> Optional[np.ndarray]:
         return self._label_ratios
 
     def switch_imbalance_status(self, flag: bool) -> None:
@@ -315,6 +368,8 @@ class ImbalancedSampler(LoggingMixin):
         if not self.shuffle or not self._sample_imbalance_flag or not self.is_imbalance:
             indices = np.arange(self._num_samples).astype(np_int_type)
         else:
+            if self._sampler is None:
+                raise ValueError("`_sampler` is not yet generated")
             indices = self._sampler.sample(self._num_samples)
         if self.shuffle:
             np.random.shuffle(indices)
@@ -326,30 +381,22 @@ class ImbalancedSampler(LoggingMixin):
         aggregation_config = None
         if self._aggregation_config is not None:
             aggregation_config = shallow_copy_dict(self._aggregation_config)
+        sample_weights = None
+        if self.sample_weights is not None:
+            sample_weights = self.sample_weights.copy()
         return ImbalancedSampler(
             self.data,
             self.imbalance_threshold,
             shuffle=self.shuffle,
             aggregation=self._aggregation_name,
             aggregation_config=aggregation_config,
+            sample_weights=sample_weights,
             sample_method=self._sample_method,
             verbose_level=self._verbose_level,
             verbose_imbalance=False,
         )
 
 
-class LabelCollators:
-    @staticmethod
-    def reg_default(y_batch):
-        assert len(y_batch) == 2
-        return y_batch[1] - y_batch[0]
-
-    @staticmethod
-    def clf_default(y_batch):
-        assert len(y_batch) == 2
-        return y_batch[1] == y_batch[0]
-
-
 class DataLoader:
     """
     Util class which can generated batches from `ImbalancedSampler`
@@ -381,38 +428,56 @@ class DataLoader:
 
     """
 
-    def __init__(self,
-                 batch_size: int,
-                 sampler: ImbalancedSampler,
-                 *,
-                 num_siamese: int = 1,
-                 return_indices: bool = False,
-                 label_collator: callable = None,
-                 verbose_level: int = 2):
+    def __init__(
+        self,
+        batch_size: int,
+        sampler: ImbalancedSampler,
+        *,
+        num_siamese: int = 1,
+        return_indices: bool = False,
+        label_collator: Optional[Callable[[np.ndarray], np.ndarray]] = None,
+        verbose_level: int = 2,
+    ):
+        self._cursor: int
+        self._siamese_cursor: int
         self._indices_in_use = None
         self._verbose_level = verbose_level
         self.data = sampler.data
+        self.sampler = sampler
         self.return_indices = return_indices
+        self._label_collator = label_collator
         if return_indices and num_siamese > 1:
-            print(f"{LoggingMixin.warning_prefix}`return_indices` is set to False because siamese loader is used")
+            print(
+                f"{LoggingMixin.warning_prefix}`return_indices` "
+                "is set to False because siamese loader is used"
+            )
             self.return_indices = False
-        self._num_siamese, self._label_collator = num_siamese, label_collator
-        self._num_samples, self.sampler = len(sampler), sampler
+        self._num_siamese = num_siamese
+        self._num_samples = len(sampler)
         self.batch_size = min(self._num_samples, batch_size)
-        self._verbose_level = verbose_level
 
-    def __len__(self):
+    def __len__(self) -> int:
         n_iter = int(self._num_samples / self.batch_size)
         return n_iter + int(n_iter * self.batch_size < self._num_samples)
 
-    def __iter__(self):
+    def __iter__(self) -> "DataLoader":
         self._reset()
         return self
 
-    def __next__(self):
+    def __next__(self) -> batch_type:
         data_next = self._get_next_batch()
         if self._num_siamese == 1:
-            return data_next
+            if self.return_indices:
+                (x_batch, y_batch), indices = data_next
+            else:
+                indices = None
+                x_batch, y_batch = data_next
+            if self._label_collator is not None:
+                y_batch = self._label_collator(y_batch)
+            batch = x_batch, y_batch
+            if not self.return_indices:
+                return batch
+            return batch, indices
         all_data = [data_next] if self._check_full_batch(data_next) else []
         while len(all_data) < self._num_siamese:
             data_next = self._get_next_batch()
@@ -428,18 +493,19 @@ class DataLoader:
         return self.sampler.sample_imbalance
 
     @enabled_sampling.setter
-    def enabled_sampling(self, value: bool):
+    def enabled_sampling(self, value: bool) -> None:
         self.sampler.switch_imbalance_status(value)
 
-    def _reset(self):
+    def _reset(self) -> None:
         reset_caches = {
             "_indices_in_use": self.sampler.get_indices(),
-            "_siamese_cursor": 0, "_cursor": -1
+            "_siamese_cursor": 0,
+            "_cursor": -1,
         }
         for attr, init_value in reset_caches.items():
             setattr(self, attr, init_value)
 
-    def _get_next_batch(self):
+    def _get_next_batch(self) -> batch_type:
         n_iter, self._cursor = len(self), self._cursor + 1
         if self._cursor == n_iter * self._num_siamese:
             raise StopIteration
@@ -448,6 +514,8 @@ class DataLoader:
             if new_siamese_cursor > self._siamese_cursor:
                 self._siamese_cursor = new_siamese_cursor
                 self._indices_in_use = self.sampler.get_indices()
+        if self._indices_in_use is None:
+            raise ValueError("`_indices_in_use` is not yet generated")
         start = (self._cursor - n_iter * self._siamese_cursor) * self.batch_size
         end = start + self.batch_size
         indices = self._indices_in_use[start:end]
@@ -456,8 +524,8 @@ class DataLoader:
             return batch
         return batch, indices
 
-    def _check_full_batch(self, batch):
-        if len(batch[0]) == self.batch_size:
+    def _check_full_batch(self, data_item: data_item_type) -> bool:
+        if len(data_item[0]) == self.batch_size:
             return True
         return False
 
@@ -474,20 +542,25 @@ class DataLoader:
 
 # time series
 
+
 class TimeSeriesModifier:
     id_name = "id"
     time_name = "timestamp"
     label_name = "label"
 
-    def __init__(self,
-                 file_path: str,
-                 task_type: TaskTypes,
-                 *,
-                 delim: str = None,
-                 ts_config: TimeSeriesConfig = None,
-                 contains_labels: bool = False):
+    def __init__(
+        self,
+        file_path: str,
+        task_type: task_type_type,
+        *,
+        delim: Optional[str] = None,
+        ts_config: Optional[TimeSeriesConfig] = None,
+        contains_labels: bool = False,
+    ):
+        task_type = parse_task_type(task_type)
         if not task_type.is_ts:
-            raise ValueError("task_type should be either TIME_SERIES_CLF or TIME_SERIES_REG")
+            msg = "task_type should be either TIME_SERIES_CLF or TIME_SERIES_REG"
+            raise ValueError(msg)
         self._file_path = file_path
         self._task_type = task_type
         self._ts_config = ts_config
@@ -496,11 +569,28 @@ class TimeSeriesModifier:
         self.data.read(file_path, contains_labels=contains_labels)
         if delim is None:
             delim = self.data._delim
+            if delim is None:
+                msg = "`data` need to contain `_delim` for `TimeSeriesModifier`"
+                raise ValueError(msg)
         self.delim = delim
-        column_names = [self.data._column_names[i] for i in sorted(self.data._column_names)]
-        self.header = None if column_names is None else self.delim.join(column_names)
-        self._num_samples = len(self.data.raw.x)
-        self.xt = self.data.raw.xT
+        if self.data._column_names is None:
+            msg = "`data` need to contain `_column_names` for `TimeSeriesModifier`"
+            raise ValueError(msg)
+        sorted_indices = sorted(self.data._column_names)
+        column_names = [self.data._column_names[i] for i in sorted_indices]
+        self.header = self.delim.join(column_names)
+        raw = self.data.raw
+        if raw is None:
+            msg = "`data` need to contain `raw` for `TimeSeriesModifier`"
+            raise ValueError(msg)
+        if raw.x is None:
+            msg = "`data` need to contain `raw.x` for `TimeSeriesModifier`"
+            raise ValueError(msg)
+        if raw.xT is None:
+            msg = "`data` need to contain `raw.xT` for `TimeSeriesModifier`"
+            raise ValueError(msg)
+        self._num_samples = len(raw.x)
+        self.xt = raw.xT
         self._modified = False
 
     def pad_id(self) -> "TimeSeriesModifier":
@@ -521,26 +611,28 @@ class TimeSeriesModifier:
         self.pad_date().pad_id()
         return self
 
-    def pad_labels(self,
-                   fn: Callable[[np.ndarray], np.ndarray],
-                   *,
-                   offset: int,
-                   batch_size: int = 32,
-                   ts_config: TimeSeriesConfig = None,
-                   aggregation_config: Dict[str, Any] = None,
-                   aggregation: str = "continuous") -> "TimeSeriesModifier":
+    def pad_labels(
+        self,
+        fn: Callable[[np.ndarray], np.ndarray],
+        *,
+        offset: int,
+        batch_size: int = 32,
+        ts_config: Optional[TimeSeriesConfig] = None,
+        aggregation_config: Optional[Dict[str, Any]] = None,
+        aggregation: str = "continuous",
+    ) -> "TimeSeriesModifier":
         if self._contains_labels:
             raise ValueError("labels already exist")
         if not self._modified:
             data = self.data
             if not data.is_ts:
-                raise ValueError("time_series_config is not provided for the original data")
+                msg = "time_series_config is not provided for the original data"
+                raise ValueError(msg)
         else:
-            column_names = None
-            if self.header is not None:
-                column_names = self.header.split(self.delim)
+            column_names = self.header.split(self.delim)
             if ts_config is None:
-                raise ValueError("columns are modified but new time_series_config is not provided")
+                msg = "columns are modified but new time_series_config is not provided"
+                raise ValueError(msg)
             data = TabularData.simple(
                 self._task_type,
                 time_series_config=ts_config,
@@ -566,8 +658,9 @@ class TimeSeriesModifier:
         self.xt.append(sorted_labels.astype(np.str))
         return self
 
-    def export_to(self,
-                  export_path: str) -> "TimeSeriesModifier":
+    def export_to(self, export_path: str) -> "TimeSeriesModifier":
+        if self.delim is None:
+            raise ValueError("`delim` is not yet generated")
         with open(export_path, "w") as f:
             if self.header is not None:
                 f.write(f"{self.header}\n")
@@ -579,84 +672,124 @@ aggregation_dict: Dict[str, Type["AggregationBase"]] = {}
 
 
 class AggregationBase(LoggingMixin, metaclass=ABCMeta):
-    def __init__(self,
-                 data: TabularData,
-                 config: Dict[str, Any],
-                 verbose_level: int):
+    def __init__(self, data: TabularData, config: Dict[str, Any], verbose_level: int):
         if not data.is_ts:
             raise ValueError("time series data is required")
         self.data = data
-        self.config, self._verbose_level = config, verbose_level
+        self.config = config
+        self._verbose_level = verbose_level
         self._num_history = config.setdefault("num_history", 1)
-        id_column = data.raw.xT[data.ts_config.id_column_idx]
+        raw = data.raw
+        if raw is None:
+            raise ValueError("`data` need to contain `raw` for `AggregationBase`")
+        if raw.xT is None:
+            raise ValueError("`data` need to contain `raw.xT` for `AggregationBase`")
+        if data.ts_config is None:
+            raise ValueError("`data` need to contain `ts_config` for `AggregationBase`")
+        id_column_idx = data.ts_config.id_column_idx
+        if id_column_idx is None:
+            msg = "`ts_config` need to contain `id_column_idx` for `AggregationBase`"
+            raise ValueError(msg)
+        id_column = raw.xT[id_column_idx]
         sorted_id_column = [id_column[i] for i in data.ts_sorting_indices]
         unique_indices = get_unique_indices(sorted_id_column)
-        self._unique_id_arr, self._id2indices = unique_indices.unique, unique_indices.split_indices
+        self.indices2id: np.ndarray
+        self._unique_id_arr = unique_indices.unique
+        self._id2indices = unique_indices.split_indices
         self._initialize()
 
     @property
     @abstractmethod
-    def num_aggregation(self):
+    def num_aggregation(self) -> int:
         pass
 
     @abstractmethod
     def _aggregate_core(self, indices: np.ndarray) -> np.ndarray:
         """ indices should be a column vector """
 
-    def _initialize(self):
-        self._num_samples_per_id = np.array(list(map(len, self._id2indices)), np_int_type)
+    def _initialize(self) -> None:
+        num_list = list(map(len, self._id2indices))
+        self._num_samples_per_id = np.array(num_list, np_int_type)
         self.log_msg("generating valid aggregation info", self.info_prefix, 5)
         valid_mask = self._num_samples_per_id >= self._num_history
         if not valid_mask.any():
             raise ValueError(
-                "current settings lead to empty valid dataset, increasing raw dataset size or "
-                f"decreasing n_history (current: {self._num_history}) might help"
+                "current settings lead to empty valid dataset, "
+                "increasing raw dataset size or decreasing n_history "
+                f"(current: {self._num_history}) might help"
             )
         if not valid_mask.all():
             invalid_mask = ~valid_mask
             n_invalid_id = invalid_mask.sum()
             n_invalid_samples = self._num_samples_per_id[invalid_mask].sum()
             self.log_msg(
-                f"{n_invalid_id} id (with {n_invalid_samples} samples) will be dropped "
-                f"(n_history={self._num_history})", self.info_prefix, verbose_level=2
+                f"{n_invalid_id} id (with {n_invalid_samples} samples) "
+                f"will be dropped (n_history={self._num_history})",
+                self.info_prefix,
+                verbose_level=2,
             )
+            invalid_ids = self._unique_id_arr[invalid_mask].tolist()
             self.log_msg(
-                f"dropped id : {', '.join(map(str, self._unique_id_arr[invalid_mask].tolist()))}",
-                self.info_prefix, verbose_level=4
+                f"dropped id : {', '.join(map(str, invalid_ids))}",
+                self.info_prefix,
+                verbose_level=4,
             )
-        self._num_samples_per_id_cumsum = np.hstack([[0], np.cumsum(self._num_samples_per_id[:-1])])
+        self._num_samples_per_id_cumsum = np.hstack(
+            [[0], np.cumsum(self._num_samples_per_id[:-1])]
+        )
         # self._id2indices need to contain 'redundant' indices here because
         # aggregation need to aggregate those 'invalid' samples
         self._id2indices_stack = np.hstack(self._id2indices)
-        self.log_msg("generating aggregation attributes", self.info_prefix, verbose_level=5)
+        self.log_msg(
+            "generating aggregation attributes",
+            self.info_prefix,
+            verbose_level=5,
+        )
         self._get_id2valid_indices()
-        self._get_valid_samples_info()
+        self._inject_valid_samples_info()
 
-    def _get_valid_samples_info(self):
+    def _inject_valid_samples_info(self) -> None:
         # 'indices' in self.indices2id here doesn't refer to indices of original dataset
         # (e.g. 'indices' in self._id2indices), but refers to indices generated by sampler,
         # so we should only care 'valid' indices here
-        self._num_valid_samples_per_id = [len(indices) for indices in self._id2valid_indices]
-        self._num_valid_samples_per_id_cumsum = np.hstack([[0], np.cumsum(self._num_valid_samples_per_id[:-1])])
-        self._num_valid_samples_per_id_cumsum = self._num_valid_samples_per_id_cumsum.astype(np_int_type)
-        self.indices2id = np.repeat(np.arange(len(self._unique_id_arr)), self._num_valid_samples_per_id)
+        self._num_valid_samples_per_id = list(map(len, self._id2valid_indices))
+        self._num_valid_samples_per_id_cumsum = np.hstack(
+            [[0], np.cumsum(self._num_valid_samples_per_id[:-1])]
+        )
+        num_int = self._num_valid_samples_per_id_cumsum.astype(np_int_type)
+        self._num_valid_samples_per_id_cumsum = num_int
+        arange = np.arange(len(self._unique_id_arr))
+        self.indices2id = np.repeat(arange, self._num_valid_samples_per_id)
         self._id2valid_indices_stack = np.hstack(self._id2valid_indices)
 
-    def _get_id2valid_indices(self):
+    def _get_id2valid_indices(self) -> None:
         # TODO : support nan_fill here
-        nan_fill, nan_ratio = map(self.config.setdefault, ["nan_fill", "nan_ratio"], ["past", 0.])
+        nan_fill = self.config.setdefault("nan_fill", "past")
+        nan_ratio = self.config.setdefault("nan_ratio", 0.0)
         self._id2valid_indices = [
-            np.array([], np_int_type) if len(indices) < self._num_history else
-            np.arange(cumsum, cumsum + len(indices) - self._num_history + 1).astype(np_int_type)
-            for cumsum, indices in zip(self._num_samples_per_id_cumsum, self._id2indices)
+            np.array([], np_int_type)
+            if len(indices) < self._num_history
+            else np.arange(
+                cumsum, cumsum + len(indices) - self._num_history + 1
+            ).astype(np_int_type)
+            for cumsum, indices in zip(
+                self._num_samples_per_id_cumsum, self._id2indices
+            )
         ]
-        self._get_valid_samples_info()
-        x, y = self.data.processed.xy
+        self._inject_valid_samples_info()
+        processed = self.data.processed
+        if processed is None:
+            raise ValueError("`processed` is not generated yet")
+        x, y = processed.xy
+        assert isinstance(x, np.ndarray)
         feature_dim = self.data.processed_dim
         for i, valid_indices in enumerate(self._id2valid_indices):
             cumsum = self._num_valid_samples_per_id_cumsum[i]
-            aggregated_flat_indices = self.aggregate(np.arange(cumsum, cumsum + len(valid_indices))).ravel()
-            aggregated_x = x[aggregated_flat_indices].reshape([-1, self.num_aggregation, feature_dim])
+            arange = np.arange(cumsum, cumsum + len(valid_indices))
+            aggregated_flat_indices = self.aggregate(arange).ravel()
+            aggregated_x = x[aggregated_flat_indices]
+            shape = [-1, self.num_aggregation, feature_dim]
+            aggregated_x = aggregated_x.reshape(shape)
             aggregated_x_nan_mask = np.isnan(aggregated_x)
             if y is None:
                 aggregated_y_valid_mask = None
@@ -674,10 +807,12 @@ class AggregationBase(LoggingMixin, metaclass=ABCMeta):
         """
         We've maintained two groups of indices in `_initialize` method:
         * the 'original' indices, which points to indices of original dataset
-        * the 'valid' indices, which is 'virtual' should points to indices of the 'original' indices
+        * the 'valid' indices, which is 'virtual' should points to indices
+        of the 'original' indices
 
-        So we need to translate sampler indices to 'valid' indices, add offsets to the 'valid' indices,
-        and then fetch the 'original' indices to fetch the corresponding data
+        So we need to translate sampler indices to 'valid' indices, add offsets
+        to the 'valid' indices, and then fetch the 'original' indices to fetch
+        the corresponding data
         * _aggregate_core method will add offsets for us
 
         Parameters
@@ -695,24 +830,24 @@ class AggregationBase(LoggingMixin, metaclass=ABCMeta):
         reversed_aggregated = self.data.ts_sorting_indices[aggregated]
         return reversed_aggregated.reshape([-1, self.num_aggregation])
 
-    def get_last_indices(self, aggregated_flat_indices: np.ndarray):
-        aggregated_indices_mat = aggregated_flat_indices.reshape([-1, self.num_aggregation])
-        return aggregated_indices_mat[..., -1]
+    def get_last_indices(self, aggregated_flat_indices: np.ndarray) -> np.ndarray:
+        reshaped = aggregated_flat_indices.reshape([-1, self.num_aggregation])
+        return reshaped[..., -1]
 
     @classmethod
-    def register(cls, name: str):
+    def register(cls, name: str) -> Callable[[Type], Type]:
         global aggregation_dict
         return register_core(name, aggregation_dict)
 
 
 @AggregationBase.register("continuous")
 class ContinuousAggregation(AggregationBase):
-    def _initialize(self):
+    def _initialize(self) -> None:
         self._history_arange = np.arange(self._num_history)
         super()._initialize()
 
     @property
-    def num_aggregation(self):
+    def num_aggregation(self) -> int:
         return self._num_history
 
     def _aggregate_core(self, indices: np.ndarray) -> np.ndarray:
@@ -720,7 +855,12 @@ class ContinuousAggregation(AggregationBase):
 
 
 __all__ = [
-    "KFold", "KRandom", "KBootstrap",
-    "ImbalancedSampler", "LabelCollators", "DataLoader",
-    "TimeSeriesModifier", "aggregation_dict", "AggregationBase",
+    "KFold",
+    "KRandom",
+    "KBootstrap",
+    "ImbalancedSampler",
+    "DataLoader",
+    "TimeSeriesModifier",
+    "aggregation_dict",
+    "AggregationBase",
 ]
